import sys

import numpy as np


def _roundup(x, multiple):
    """ Round up `x` to multiple of `multiple`"""
    return int((x + multiple - 1) // multiple * multiple)


def _rounddown(x, multiple):
    """ Round down `x` to multiple of `multiple`"""
    return int(x // multiple * multiple)


def _is_batch_full(num_tokens, max_tokens, max_sentences, batch_length):
    """Returns true if batch is full

    Args:
        num_tokens (int): number of tokens in batch
        max_tokens (Optional[int]): Max number of tokens in batch
        max_sentences (Optional[int]): Max number of sentences in batch
        batch_length (int): Current batch length

    Returns:
        (bool): Whether the batch is full or not
    """
    if batch_length == 0 or (max_tokens is None and max_sentences is None):
        return False

    elif batch_length == max_sentences or num_tokens > max_tokens:
        return True
    else:
        return False


def _make_batches(
    src_lengths,
    trg_lengths,
    indices,
    max_tokens,
    max_sentences,
    max_len,
    bsz_mult,
    pad_seq,
):
    """Creates the batches for WMT17 Dataset

    Args:
        src_lengths (iterable[int]): The source lengths
        trg_lengths (iterable[int]): The target lengths
        indices (iterable[int]): Indices to consider (must be sorted by lengths)
        max_tokens (Optional[int]): Max tokens per batch
        max_sentences (optional[int]): Max sentences per batch
        max_len (int): Maximum sequence length
        bsz_mult (int): Batch size multiple
        pad_seq (int): Multiple to round up to for sequence lengths

    Returns:
        (list[list[int]]): The list of batches, referenced by index
    """
    batches = []
    nelem = len(indices)
    sample_len = 0

    num_seqs_mult = bsz_mult // pad_seq if (bsz_mult % pad_seq == 0) else bsz_mult

    sample_lens = []
    batch = []
    for i in range(nelem):
        idx = indices[i]
        sample_num_tokens = max(src_lengths[idx], trg_lengths[idx])
        if sample_num_tokens > max_len:
            continue

        sample_len = max(sample_len, sample_num_tokens)
        padded_sample_len = (
            _roundup(sample_len, bsz_mult)
            if len(batch) < num_seqs_mult
            else _roundup(sample_len, pad_seq)
        )
        sample_lens.append(sample_num_tokens)
        num_tokens = (len(batch) + 1) * padded_sample_len

        if _is_batch_full(num_tokens, max_tokens, max_sentences, len(batch)):
            sequences = len(batch)
            if (sequences % num_seqs_mult != 0) and (sequences > num_seqs_mult):
                pad_sequences_opt_seqs = _rounddown(sequences, num_seqs_mult)
                total_tokens_opt_seqs = padded_sample_len * pad_sequences_opt_seqs

                pad_seq_len_opt_seqlen = _roundup(padded_sample_len, bsz_mult)
                pad_sequences_opt_seqlen = max_tokens // pad_seq_len_opt_seqlen
                total_tokens_opt_seqlen = padded_sample_len * pad_sequences_opt_seqlen

                if total_tokens_opt_seqs >= total_tokens_opt_seqlen:
                    sequences = pad_sequences_opt_seqs
                else:
                    sequences = pad_sequences_opt_seqlen

            # Copy overflowing data to next batch if sequences < len(batch)
            new_batch = batch[sequences:]
            batch = batch[:sequences]
            sample_lens = sample_lens[sequences:]  # keep lens for next iter
            sample_len = max(sample_lens)
            batches.append(batch)
            batch = new_batch

        batch.append(idx)

    while len(batch) > 0:
        sequences = max(
            _rounddown(len(batch), num_seqs_mult), len(batch) % num_seqs_mult
        )
        new_batch = batch[sequences:]
        batch = batch[:sequences]
        batches.append(batch)
        batch = new_batch

    return batches


def get_batches(
    dataset, max_tokens=None, max_sentences=None, bsz_mult=8, shuffle=True, seed=0
):
    """Creates the batches for the given dataset

    Args:
        dataset (`obj`:torch.utils.Dataset): The dataset
        max_tokens (Optional[int]): Maximum number of tokens in one batch.
            Default: `None` (no limit)
        max_sentences (int): Maximum number of sentences per batch.
            Default: `None` (no limit)
        bsz_mult (int): Batch size multiple. Default: 8
        shuffle (bool): Shuffle batches. Default: `True`
        seed (int): Seed to use for shuffling. Default: 0

    Returns:
        (list[list[int]]): The batches
    """
    if hasattr(dataset, "indices"):
        partition_indices = dataset.indices
    else:
        partition_indices = None
    indices = dataset.ordered_indices(partition_indices, seed=seed)
    src_sizes = dataset.src_sizes
    trg_sizes = dataset.trg_sizes

    max_sentences = max_sentences if max_sentences is not None else sys.maxsize
    max_tokens = max_tokens if max_tokens is not None else sys.maxsize
    max_pos_num = min(dataset.max_source_positions, dataset.max_target_positions)

    batches = _make_batches(
        src_sizes,
        trg_sizes,
        indices,
        max_tokens,
        max_sentences,
        max_pos_num,
        bsz_mult,
        dataset.seq_len_multiple,
    )

    if shuffle:
        np.random.seed(seed)
        np.random.shuffle(batches)
    return batches



import logging

import torch
from torch.nn.utils import clip_grad_norm_

from mlbench_core.utils.pytorch.distributed import (
    AllReduceAggregation,
    AllReduceAggregationHVD,
)

try:
    from apex import amp
except ImportError as e:
    pass

logger = logging.getLogger("mlbench")


class DynamicLossScaler:
    def __init__(
        self, init_scale=2.0 ** 15, scale_factor=2.0, scale_window=2000, max_scale=None
    ):
        self.loss_scale = init_scale
        self.scale_factor = scale_factor
        self.scale_window = scale_window
        self._iter = 0
        self._last_overflow_iter = -1
        self.max_scale = max_scale

    def update_scale(self, overflow):
        if overflow:
            self.loss_scale /= self.scale_factor
            self._last_overflow_iter = self._iter
        elif (self._iter - self._last_overflow_iter) % self.scale_window == 0:
            self.loss_scale *= self.scale_factor
            if self.max_scale is not None:
                self.loss_scale = min(self.loss_scale, self.max_scale)
        self._iter += 1

    @staticmethod
    def has_overflow(grad_norm):
        # detect inf and nan
        if grad_norm == float("inf") or grad_norm != grad_norm:
            return True
        return False


class FP16Optimizer:
    """
    Mixed precision optimizer with dynamic loss scaling and backoff.
    https://docs.nvidia.com/deeplearning/sdk/mixed-precision-training/index.html#scalefactor

    Args:
        fp16_model (`obj`:torch.nn.Module): model (previously casted to half)
        world_size (int): Distributed world size
        use_cuda (bool): Use cuda tensors for aggregation
        use_horovod (bool): Use Horovod for aggregation
        by_layer (bool): Aggregate by layer
        grad_clip (float): coefficient for gradient clipping, max L2 norm of the gradients
        init_scale (int):  initial loss scale
        scale_factor (float): Factor for upscale/dowscale
        scale_window (int): interval for loss scale upscaling
        average_world (bool): Average the gradients by world size
        average_custom (bool): Divide gradients by given denominator at each step, instead
            of `world_size`
        divide_before (bool): Divide gradients before reduction (default: False)
    """

    def __init__(
        self,
        fp16_model,
        world_size,
        use_cuda=False,
        use_horovod=False,
        by_layer=False,
        grad_clip=float("inf"),
        init_scale=1024,
        scale_factor=2,
        scale_window=128,
        max_scale=None,  # TODO max scale for GNMT is 8192.0
        min_scale=1e-4,
        average_world=False,
        average_custom=False,
        divide_before=False,
    ):
        self.use_cuda = use_cuda

        self.fp16_model = fp16_model
        self.fp32_params = self.initialize_flat_fp32_weight()

        self.loss_scaler = DynamicLossScaler(
            init_scale=init_scale,
            scale_factor=scale_factor,
            scale_window=scale_window,
            max_scale=max_scale,
        )
        self.min_scale = min_scale
        self.grad_clip = grad_clip

        self.optimizer = None

        if use_horovod:
            self.agg = AllReduceAggregationHVD(
                world_size=world_size, use_cuda=use_cuda, divide_before=divide_before
            ).agg_grad(by_layer=by_layer)
        else:
            self.agg = AllReduceAggregation(
                world_size=world_size, use_cuda=use_cuda, divide_before=divide_before
            ).agg_grad(by_layer=by_layer)

        if average_world:
            self.agg_mode = "avg_world"
        elif average_custom:
            self.agg_mode = "custom_avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    # Flattening master weight
    def initialize_flat_fp32_weight(self):
        """ Initializes the model's parameters in fp32

        Returns:
            (`obj`:torch.Tensor): The Parameters in fp32
        """
        # Get all params that require gradient
        params = [p for p in self.fp16_model.parameters() if p.requires_grad]
        total_param_size = sum(p.data.numel() for p in params)

        # Create new fp32 params
        fp32_params = params[0].new(0).float().new(total_param_size)
        offset = 0
        for p in params:
            numel = p.data.numel()
            fp32_params[offset : offset + numel].copy_(p.data.view(-1))
            offset += numel

        fp32_params = torch.nn.Parameter(fp32_params, requires_grad=True)

        fp32_params.grad = torch.autograd.Variable(
            fp32_params.data.new(*fp32_params.size())
        )

        return fp32_params

    @staticmethod
    def fp16_to_fp32_flat_grad(fp32_params, fp16_model):
        """ Copies the parameters in `fp16_model` into `fp32_params` in-place

        Args:
            fp32_params (torch.Tensor): Parameters in fp32
            fp16_model (torch.nn.Module): Model in fp16

        """
        flat_grads = torch.cat(
            [p.grad.data.view(-1) for p in fp16_model.parameters() if p.requires_grad]
        )
        fp32_params.grad = flat_grads.to(torch.float32)

    @staticmethod
    def fp32_to_fp16_weights(fp16_model, fp32_params):
        """ Copies the parameters in `fp32_params` into `fp16_model` in-place

         Args:
             fp16_model (torch.nn.Module): Model in fp16
             fp32_params (torch.Tensor): Parameters in fp32

         """
        with torch.no_grad():
            pointer = 0
            for p in fp16_model.parameters():
                if not p.requires_grad:
                    continue
                nelem = p.numel()
                p.data.copy_(
                    fp32_params.data[pointer : pointer + nelem].view_as(p.data)
                )
                pointer += nelem

    def backward_loss(self, loss):
        """ Scales and performs backward on the given loss

        Args:
            loss (torch.nn.Module): The loss

        """
        loss *= self.loss_scaler.loss_scale
        loss.backward()

    def step(self, closure=None, multiplier=1, denom=None):
        """
        Performs one step of the optimizer.
        Applies loss scaling, computes gradients in fp16, converts gradients to
        fp32, inverts scaling and applies optional gradient norm clipping.
        If gradients are finite, it applies update to fp32 master weights and
        copies updated parameters to fp16 model for the next iteration. If
        gradients are not finite, it skips the batch and adjusts scaling factor
        for the next iteration.

        Args:
            closure (callable, optional): A closure that reevaluates the model and returns the loss.
            multiplier (float): Multiplier for gradient scaling. Gradient will be scaled using
                `scaled_grad = reduced_grad / (loss_scaler * multiplier)`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with `average_batch`. (default: `None`)
        """

        scaling_factor = self.loss_scaler.loss_scale * multiplier

        # Aggregate gradients
        self.agg(self.fp16_model, self.agg_mode, denom=denom)

        # Cast fp16 grads to fp32 for optimizer
        self.fp16_to_fp32_flat_grad(self.fp32_params, self.fp16_model)

        # UnScale gradients
        if scaling_factor != 1.0:
            self.fp32_params.grad.data.div_(scaling_factor)

        # Clip and compute gradient norm
        norm = clip_grad_norm_([self.fp32_params], self.grad_clip)
        updated = False
        overflow = self.loss_scaler.has_overflow(norm)
        self.loss_scaler.update_scale(overflow)

        if not overflow:
            self.optimizer.step(closure=closure)
            self.fp32_to_fp16_weights(self.fp16_model, self.fp32_params)
            updated = True
        else:
            if self.loss_scaler.loss_scale <= self.min_scale:
                raise Exception(
                    "Minimum loss scale ({}) reached".format(self.min_scale)
                )
            logger.info(f"Skipped batch, new scale: {self.loss_scaler.loss_scale}")

        return updated

    def zero_grad(self):
        """Resets the gradients of the optimizer and fp16_model"""
        self.optimizer.zero_grad()
        self.fp16_model.zero_grad()


class FP32Optimizer:
    """
    Standard optimizer, computes backward and applies weight update.

    Args:
        model (`obj`:torch.nn.Module): model
        world_size (int): Distributed world size
        use_cuda (bool): Use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer
        grad_clip (float): coefficient for gradient clipping, max L2 norm of the gradients
        average_world (bool): Average the gradients by world size
        average_custom (bool): Divide gradients by given denominator at each step, instead
            of `world_size`
        divide_before (bool): Divide gradients before reduction (default: False)
    """

    def __init__(
        self,
        model,
        world_size,
        use_cuda=False,
        by_layer=False,
        grad_clip=float("inf"),
        average_world=False,
        average_custom=False,
        divide_before=False,
    ):
        self.model = model
        self.grad_clip = grad_clip
        self.optimizer = None

        if average_world:
            self.agg_mode = "avg_world"
        elif average_custom:
            self.agg_mode = "custom_avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        self.agg = AllReduceAggregation(
            world_size=world_size, use_cuda=use_cuda, divide_before=divide_before
        ).agg_grad(by_layer=by_layer)

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def step(self, closure=None, denom=None):
        """
        Performs one step of the optimizer.

        Args:
            closure (callable): Optimizer closure argument
            denom (:obj:`torch.Tensor`, optional): Custom denominator to reduce by
        """
        # Aggregate gradients
        self.agg(self.model, self.agg_mode, denom=denom)

        # Clip norm
        if self.grad_clip != float("inf"):
            clip_grad_norm_(self.model.parameters(), self.grad_clip)

        # Optimizer step
        self.optimizer.step(closure=closure)
        return True

    def backward_loss(self, loss):
        loss.backward()

    def zero_grad(self):
        self.optimizer.zero_grad()


class AMPOptimizer:
    """
    Optimizer compatible with AMP.
    Uses AMP to apply loss scaling, computes backward and applies weight
    update.

    Args:
        model (`obj`:torch.nn.Module): model
        grad_clip (float): coefficient for gradient clipping, max L2 norm of the gradients
        loss_scale (int):  initial loss scale
        dls_upscale_interval (int): interval for loss scale upscaling
        average_world (bool): Average the gradients by world size
        average_custom (bool): Divide gradients by given denominator at each step, instead
            of `world_size`
        world_size (int): Distributed world size
        use_cuda (bool): Use cuda tensors for aggregation
        by_layer (bool): Aggregate by layer
        use_horovod (bool): Use Horovod for aggregation
    """

    def __init__(
        self,
        model,
        grad_clip=None,
        loss_scale=8192,
        dls_upscale_interval=128,
        average_world=True,
        average_custom=False,
        world_size=1,
        use_cuda=False,
        by_layer=False,
        use_horovod=False,
    ):
        self.model = model
        self.grad_clip = grad_clip
        self.optimizer = None
        loss_scaler = amp._amp_state.loss_scalers[0]
        loss_scaler._loss_scale = loss_scale
        loss_scaler._scale_seq_len = dls_upscale_interval

        if average_world:
            self.agg_mode = "avg_world"
        elif average_custom:
            self.agg_mode = "custom_avg"
        else:
            raise NotImplementedError("Only average model is supported right now.")

        if use_horovod:
            self.agg = AllReduceAggregationHVD(
                world_size=world_size, use_cuda=use_cuda
            ).agg_grad(by_layer=by_layer)
        else:
            self.agg = AllReduceAggregation(
                world_size=world_size, use_cuda=use_cuda
            ).agg_grad(by_layer=by_layer)

    def set_optimizer(self, optimizer):
        self.optimizer = optimizer

    def backward_loss(self, loss):
        with amp.scale_loss(loss, self.optimizer) as scaled_loss:
            scaled_loss.backward()

    def step(self, closure=None):
        """
        Performs one step of the optimizer.
        """
        if self.grad_clip != float("inf"):
            clip_grad_norm_(amp.master_params(self.optimizer), self.grad_clip)

        self.agg(self.model, self.agg_mode)
        self.optimizer.step(closure=closure)
        return True

    def zero_grad(self):
        self.optimizer.zero_grad()



import struct

import numpy as np

dtypes = {
    1: np.uint8,
    2: np.int8,
    3: np.int16,
    4: np.int32,
    5: np.int64,
    6: np.float,
    7: np.double,
}


def write_longs(f, a):
    f.write(np.array(a, dtype=np.int64))


def code(dtype):
    for k in dtypes.keys():
        if dtypes[k] == dtype:
            return k


class IndexedDatasetBuilder(object):
    element_sizes = {
        np.uint8: 1,
        np.int8: 1,
        np.int16: 2,
        np.int32: 4,
        np.int64: 8,
        np.float: 4,
        np.double: 8,
    }

    def __init__(self, out_file, dtype=np.int32):
        self.out_file = open(out_file, "wb")
        self.dtype = dtype
        self.data_offsets = [0]
        self.dim_offsets = [0]
        self.sizes = []
        self.element_size = self.element_sizes[self.dtype]

    def add_item(self, tensor):
        bytes = self.out_file.write(np.array(tensor.numpy(), dtype=self.dtype))
        self.data_offsets.append(self.data_offsets[-1] + bytes / self.element_size)
        for s in tensor.size():
            self.sizes.append(s)
        self.dim_offsets.append(self.dim_offsets[-1] + len(tensor.size()))

    def finalize(self, index_file):
        self.out_file.close()
        index = open(index_file, "wb")
        index.write(b"TNTIDX\x00\x00")
        index.write(struct.pack("<Q", 1))
        index.write(struct.pack("<QQ", code(self.dtype), self.element_size))
        index.write(struct.pack("<QQ", len(self.data_offsets) - 1, len(self.sizes)))
        write_longs(index, self.dim_offsets)
        write_longs(index, self.data_offsets)
        write_longs(index, self.sizes)
        index.close()



import torch


def _collate_tokens(
    values,
    pad_idx,
    eos_idx,
    left_pad,
    move_eos_to_beginning=False,
    n_seq_per_batch_multiple=8,
    seq_len_multiple=1,
):
    """ Convert a list of 1d tensors into a padded 2d tensor.

    Args:
        values (list[torch.Tensor]): A list of tensors
        pad_idx (int): Padding symbol index
        eos_idx (int): EOS symbol index
        left_pad (bool): left- or right-padding (true: left, false: right)
        move_eos_to_beginning (bool): Reverse order of sequence of tokens (true: reverse, false: original)
        n_seq_per_batch_multiple (int): The number of sequences per batch to round down to
        seq_len_multiple (int): The number of tokens per sequence to round up to

    Returns:
        (`obj`:torch.Tensor): The tensor of collated and padded tokens
    """
    size_of_seq_dim = max(v.size(0) for v in values)  # Unpadded size
    n_seq_in_batch = len(values)

    if n_seq_per_batch_multiple % seq_len_multiple == 0:
        n_seq_multiple = n_seq_per_batch_multiple / seq_len_multiple
    else:
        n_seq_multiple = n_seq_per_batch_multiple

    if n_seq_in_batch < n_seq_multiple or n_seq_in_batch % n_seq_multiple > 0:
        seq_len_multiple = n_seq_per_batch_multiple

    size_of_seq_dim = (
        (size_of_seq_dim + seq_len_multiple - 1) // seq_len_multiple * seq_len_multiple
    )  # Padded seq len, rounded up to next multiple

    padded_2d_tensor = values[0].new(len(values), size_of_seq_dim).fill_(pad_idx)

    def copy_tensor(src, dst):
        assert dst.numel() == src.numel()

        if move_eos_to_beginning:
            assert src[-1] == eos_idx
            dst[0] = eos_idx
            dst[1:] = src[:-1]
        else:
            dst.copy_(src)

    if left_pad:
        for idx, val in enumerate(values):
            copy_tensor(val, padded_2d_tensor[idx][size_of_seq_dim - len(val) :])
    else:
        for idx, val in enumerate(values):
            copy_tensor(val, padded_2d_tensor[idx][: len(val)])

    return padded_2d_tensor


def collate_batch(
    samples,
    pad_idx,
    eos_idx,
    left_pad_source=True,
    left_pad_target=False,
    bsz_mult=8,
    seq_len_multiple=1,
):
    """Collate a list of samples into a batch

    Args:
        samples (list[dict]): Samples to collate
        pad_idx (int): Padding symbol index
        eos_idx (int): EOS symbol index
        left_pad_source (bool): Pad sources on the left
        left_pad_target (bool): Pad sources on the right
        bsz_mult (int): Batch size multiple
        seq_len_multiple (int): Sequence length multiple

    Returns:
        (dict): `{'id' (list[int]): list of indexes,
                  'ntokens' (int): total number of tokens,
                  'net_input' (dict): input of net, containing 'src_tokens', 'src_lengths'
                                      and 'prev_output_tokens', all tensors,
                  'target' (Tensor): Target Tensor}`

    """
    if len(samples) == 0:
        return {}

    def merge(key, left_pad, move_eos_to_beginning=False):
        return _collate_tokens(
            [s[key] for s in samples],
            pad_idx,
            eos_idx,
            left_pad,
            move_eos_to_beginning,
            bsz_mult,
            seq_len_multiple,
        )

    id = torch.LongTensor([s["id"] for s in samples])
    src_tokens = merge("source", left_pad=left_pad_source)
    # sort by descending source length
    src_lengths = torch.LongTensor([s["source"].numel() for s in samples])

    prev_output_tokens = None
    target = None
    if samples[0].get("target", None) is not None:
        target = merge("target", left_pad=left_pad_target)
        # we create a shifted version of targets for feeding the
        # previous output token(s) into the next decoder step
        prev_output_tokens = merge(
            "target", left_pad=left_pad_target, move_eos_to_beginning=True,
        )
        ntokens = sum(len(s["target"]) for s in samples)
    else:
        ntokens = sum(len(s["source"]) for s in samples)

    return {
        "id": id,
        "ntokens": ntokens,
        "net_input": {
            "src_tokens": src_tokens,
            "src_lengths": src_lengths,
            "prev_output_tokens": prev_output_tokens,
        },
        "target": target,
    }



import numpy as np
import torch
import torch.distributed as dist

from mlbench_core.utils.pytorch.utils import orthogonalize

try:
    import horovod.torch as hvd
except ImportError as e:
    pass

ALLREDUCE_AGGREGATION_OPS = ["avg_world", "custom_avg"]
"""
All possible aggregations for AllReduceAggregation
"""


def global_average(sum, count):
    def helper(array):
        array = get_backend_tensor(torch.Tensor(array))

        dist.all_reduce(array, op=dist.ReduceOp.SUM)
        return array[0] / array[1]

    avg = helper([sum, count])
    return avg


def pack_tensors(tensors, use_cuda=False):
    """
    Packs a list of tensors into one 1-dimensional tensor.

    Args:
        tensors (list[torch.Tensor]): The tensors to pack
        use_cuda (bool): Whether the resulting tensor should be on cuda

    Returns:
        (torch.Tensor, list[int], list[(int, int)]):
            The flattened tensors, the list start indices of each packed tensor,
            and the original shape of each tensor.

            Those values are used to then unpack the tensor
    """
    indices = [0]
    for tensor in tensors:
        new_end = indices[-1] + tensor.nelement()
        indices.append(new_end)

    tensor_sizes = [t.size() for t in tensors]

    vec = torch.empty(
        indices[-1],
        device=tensors[0].device if tensors[0].is_cuda and use_cuda else "cpu",
        dtype=tensors[0].dtype,
    )

    for tensor, start_idx, end_idx in zip(tensors, indices[:-1], indices[1:]):
        vec[start_idx:end_idx] = tensor.data.view(-1)

    return vec, indices, tensor_sizes


def unpack_tensors(aggregated, indices, sizes):
    """
    Unpacks a 1-dimensional tensor into a list of tensors

    Args:
        aggregated (torch.Tensor): The 1-dimensional tensor
        indices (List[Int]): The start index of each tensor
        sizes (List[(Int, Int)]): The size of each resulting tensor

    Returns:
        List[torch.Tensor]: The unpacked tensors
    """
    start_index = indices[:-1]
    end_index = indices[1:]

    tensors = []
    for i, (start, end) in enumerate(zip(start_index, end_index)):
        tensors.append(aggregated[start:end].view(sizes[i]))

    return tensors


##########################################################################################


class Aggregation(object):
    """Aggregate updates / models from different processes.

    Args:
        use_cuda (bool): Whether to use CUDA tensors for communication
    """

    def __init__(self, use_cuda=False):
        self.use_cuda = use_cuda

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggregated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        raise NotImplementedError

    def _agg_weights_by_model(self, model, op, denom=None):
        """Aggregate models by model weight, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Pack all layers
        packed, indices, sizes = pack_tensors(
            [t for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op, denom=denom)

        tensors = unpack_tensors(aggregated, indices, sizes)
        # Unpack
        for i, param in enumerate(model.parameters()):
            param.data = tensors[i]

    def _agg_gradients_by_model(self, model, op, denom=None):
        """Aggregate models gradients, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Pack all layers
        packed, indices, sizes = pack_tensors(
            [t.grad for t in model.parameters()], use_cuda=self.use_cuda
        )
        aggregated = self._agg(packed, op=op, denom=denom)

        # Unpack
        tensors = unpack_tensors(aggregated, indices, sizes)
        for i, param in enumerate(model.parameters()):
            param.grad.data = tensors[i]

    def _agg_weights_by_layer(self, model, op, denom=None):
        """Aggregate models by model weight, for each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.data, op=op, denom=denom)
            param.data = grad

    def _agg_gradients_by_layer(self, model, op, denom=None):
        """Aggregate models gradients each layer individually

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)
        """
        # Aggregate layer by layer
        for _, param in enumerate(model.parameters()):
            grad = self._agg(param.grad.data, op=op, denom=denom)
            param.grad.data = grad

    def agg_model(self, by_layer=False):
        if by_layer:
            return self._agg_weights_by_layer
        else:
            return self._agg_weights_by_model

    def agg_grad(self, by_layer=False):
        if by_layer:
            return self._agg_gradients_by_layer
        else:
            return self._agg_gradients_by_model


class AllReduceAggregation(Aggregation):
    """Aggregate udpates / models from different processes using all-reduce aggregation

    Args:
        world_size (int): Current distributed world size
        divide_before (bool): Perform division before reduction (avoid overflow)
        use_cuda (bool): Use cuda tensors for reduction
    """

    def __init__(self, world_size, divide_before=False, use_cuda=False):
        self.world_size = world_size
        self.divide_before = divide_before
        super(AllReduceAggregation, self).__init__(use_cuda=use_cuda)

    def _reduce(self, data):
        """Reduces the given tensor using `torch.distributed` and op=`dist.ReduceOp.SUM`

        Args:
            data (`obj`:torch.Tensor): The tensor to reduce

        Returns:
            (`obj`:torch.Tensor): The reduced Tensor
        """
        dist.all_reduce(data, op=dist.ReduceOp.SUM)
        return data

    def _divide(self, data, op, denom=None):
        """Divides the given `data` tensor by

        - `world_size` if op == `avg_world`
        - `denom`, if op == `custom_avg`

        Args:
            data (`obj`:torch.Tensor): Data tensor to divide
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            (`obj`:torch.Tensor): The resulting tensor
        """
        if op not in ALLREDUCE_AGGREGATION_OPS:
            raise NotImplementedError("Allreduce not implemented for op={}".format(op))

        if op == "avg_world":
            data.div_(self.world_size)
        elif op == "custom_avg":
            if denom is None or denom == 0:
                raise ValueError("Denominator should be one element tensor")
            data.div_(denom)

        return data

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        - If op == `avg_world`, the reduced tensor will be divided by `world_size`,
        - if op == `custom_avg`, the reduced tensor will be divided by `denom`

        If `self.divide_before`, the division is performed before reduction.
        This can be helpful to avoid overflows when using `float16` training

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggregated.
            op (str): Aggregation method. Should be in `ALLREDUCE_AGGREGATION_OPS`
            denom (:obj:`torch.Tensor`, optional): Custom denominator to average by
                Use with op == `custom_avg`. (default: `None`)

        Returns:
            (`obj`:torch.Tensor): The aggregated tensor.
        """
        if self.divide_before:
            data = self._divide(data, op, denom)

        if self.world_size > 1:
            data = self._reduce(data)

        if not self.divide_before:
            data = self._divide(data, op, denom)
        return data


class AllReduceAggregationHVD(AllReduceAggregation):
    def _reduce(self, data):
        data = hvd.allreduce(data, op=hvd.Sum)
        return data


class DecentralizedAggregation(Aggregation):
    """Aggregate updates in a decentralized manner."""

    def __init__(self, rank, neighbors, use_cuda=False):
        """
        Args:
            rank (int): Rank of the current process
            neighbors (list): A list of ranks of its neighbors.
        """
        assert rank not in neighbors
        self.rank = rank
        self.neighbors = neighbors
        super(DecentralizedAggregation, self).__init__(use_cuda=use_cuda)

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        # Create some tensors to host the values from neighborhood.
        local_data = {i: torch.zeros_like(data) for i in self.neighbors}
        local_data[self.rank] = data

        reqs = []
        for node in self.neighbors:
            reqs.append(dist.isend(tensor=local_data[self.rank], dst=node))
            reqs.append(dist.irecv(tensor=local_data[node], src=node))

        for req in reqs:
            req.wait()

        # Aggregate local_data
        if op == "avg_world":
            output = sum(local_data.values()) / (len(self.neighbors) + 1)
        else:
            raise NotImplementedError("op {} is not supported yet.".format(op))

        return output


class SparsifiedAggregation(Aggregation):
    """Aggregate sparsified updates."""

    def __init__(self, model, use_cuda=False):
        super(SparsifiedAggregation, self).__init__(use_cuda=use_cuda)
        pass

    def _agg(self, data, op, denom=None):
        pass


class PowerAggregation(Aggregation):
    """Aggregate updates using power iteration and error feedback.

    Args:
            model (:obj:`nn.Module`): Model which contains parameters for SGD
            use_cuda (bool): Whether to use cuda tensors for aggregation
            reuse_query (bool): Whether to use warm start to initialize the power iteration
            rank (int): The rank of the gradient approximation
    """

    def __init__(self, model, use_cuda=False, reuse_query=False, rank=1):
        super(PowerAggregation, self).__init__(use_cuda=use_cuda)
        self.p_memory = None
        self.q_memory = None
        self.reuse_query = reuse_query
        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.rng = np.random.RandomState(1)
        self.n_workers = dist.get_world_size()
        self.rank = rank
        self.memories = [torch.zeros_like(param) for param in model.parameters()]
        self.send_buffers = [torch.zeros_like(param) for param in model.parameters()]

    def set_random(self, vector):
        """Sets the data in `vector` to random values."""
        torch.manual_seed(self.rng.randint(1_000_000_000))
        vector.data[:] = torch.randn(*vector.shape, device=self.device)

    def _agg_gradients_by_model(self, model, op, denom=None):
        """Aggregate models gradients, all layers at once

        Args:
            model (:obj:`torch.Module`): Models to be averaged.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (None): Not used here

        """
        grads = [t.grad.data for t in model.parameters()]
        aggregated = self._agg(grads, op=op)

        for i, param in enumerate(model.parameters()):
            param.grad.data = aggregated[i]

    def agg_weights(self, by_layer=False):
        raise NotImplementedError("PowerSGD doesn't allow aggregation by weights")

    def agg_model(self, by_layer=False):
        if by_layer:
            raise NotImplementedError("PowerSGD doesn't allow aggregation by layer")
        else:
            return self._agg_gradients_by_model

    def _agg(self, data, op, denom=None):
        """Aggregate data using `op` operation.

        Args:
            data (:obj:`torch.Tensor`): A Tensor to be aggragated.
            op (str): Aggregation methods like `avg`, `sum`, `min`, `max`, etc.
            denom (None): Not used here

        Returns:
            :obj:`torch.Tensor`: An aggregated tensor.
        """
        if op == "avg":
            for grad, memory, send_bfr in zip(data, self.memories, self.send_buffers):
                send_bfr.data[:] = grad + memory
            self.reduce(self.send_buffers, data, self.memories)
        else:
            raise NotImplementedError("op {} is not supported yet.".format(op))
        return data

    def reduce(self, grad_in, grad_out, memory_out):
        """Reduces the gradients between the workers in place and calculates error feedback.

        Args:
            grad_in (list[torch.Tensor]): The gradients to reduce.
            grad_out (list[torch.Tensor]): Used for storing the reduced gradients.
            memory_out (list[torch.Tensor]): Used for storing error feedback.
        """
        # Split the tensors into rank1-ones that will be reduced un-compressed
        # and rank > 1 tensors that are compressed
        rank1_tensors = [
            (tensor, out, mem)
            for tensor, out, mem in zip(grad_in, grad_out, memory_out)
            if tensor.ndimension() <= 1
        ]
        high_rank_tensors = [
            (tensor, out, mem)
            for tensor, out, mem in zip(grad_in, grad_out, memory_out)
            if tensor.ndimension() > 1
        ]

        # We are building a rank-1 approximation of every tensor
        # that can be interpreted as a matrix. Let the approximation be
        # M = p q^T
        # We are allocating consequtive memory for the p's and q's

        memory_is_uninitialized = self.p_memory is None

        p_total_size = 0
        q_total_size = 0
        for tensor, _, _ in high_rank_tensors:
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape
            rank = min(n, m, self.rank)
            p_total_size += n * rank
            q_total_size += m * rank
        if self.p_memory is None:
            self.p_memory = torch.empty(p_total_size, device=self.device)
            self.q_memory = torch.empty(q_total_size, device=self.device)

        # Find them again and make lists of pointers
        ps = []
        qs = []
        p_idx = 0
        q_idx = 0
        for tensor, _, _ in high_rank_tensors:
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape
            rank = min(n, m, self.rank)
            ps.append(self.p_memory[p_idx : p_idx + n * rank].view(n, rank))
            qs.append(self.q_memory[q_idx : q_idx + m * rank].view(m, rank))
            p_idx += n * rank
            q_idx += m * rank

        for (tensor, _, _), q, p in zip(high_rank_tensors, qs, ps):
            matrix = tensor.view(tensor.shape[0], -1)
            n, m = matrix.shape

            if self.reuse_query and not memory_is_uninitialized:
                pass
            else:
                # Sample a query vector q
                self.set_random(q)

        for (tensor, _, _), q, p in zip(high_rank_tensors, qs, ps):
            matrix = tensor.view(tensor.shape[0], -1)
            torch.matmul(matrix, q, out=p)

        dist.all_reduce(self.p_memory)

        # Start communicating rank 1 tensors
        rank1_packed, rank1_indices, rank1_sizes = pack_tensors(
            [tensor for (tensor, _, _) in rank1_tensors]
        )

        rank1_handle = dist.all_reduce(rank1_packed, async_op=True)

        for p in ps:
            orthogonalize(p)

        for p, q, (tensor, _, _) in zip(ps, qs, high_rank_tensors):
            matrix = tensor.view(tensor.shape[0], -1)
            torch.matmul(matrix.t(), p, out=q)

        dist.all_reduce(self.q_memory)
        self.q_memory.data[:] /= self.n_workers

        for p, q, (tensor, out, mem) in zip(ps, qs, high_rank_tensors):
            # Set the output gradient
            torch.matmul(p, q.t(), out=out.data[:])
            mem.data[:] = tensor - out

        rank1_handle.wait()
        rank1_packed /= self.n_workers
        rank1_unpacked = unpack_tensors(rank1_packed, rank1_indices, rank1_sizes)
        for i, (_, out, _) in enumerate(rank1_tensors):
            out[:] = rank1_unpacked[i]


def get_backend_tensor(tensor):
    if dist.is_initialized() and dist.get_backend() == dist.Backend.NCCL:
        return tensor.cuda()
    return tensor



# Copyright 2018 MLBenchmark Group. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Download and preprocess WMT17 en2de training and evaluation datasets."""

from __future__ import absolute_import, division, print_function

import argparse
import glob
import os
import random
import shutil
import subprocess
import tarfile
import urllib
import urllib.request
from collections import Counter

import torch

from mlbench_core.dataset.nlp.pytorch.wmt17.preprocess import (
    indexed_dataset,
    sub_tokenizer,
)

# Data sources for training/evaluating the transformer translation model.
# If any of the training sources are changed, then either:
#   1) use the flag `--search` to find the best min count or
#   2) update the _TRAIN_DATA_MIN_COUNT constant.
# min_count is the minimum number of times a token must appear in the data
# before it is added to the vocabulary. "Best min count" refers to the value
# that generates a vocabulary set that is closest in size to _TARGET_VOCAB_SIZE.
_TRAIN_DATA_SOURCES = [
    {
        "url": "http://data.statmt.org/wmt17/translation-task/"
        "training-parallel-nc-v12.tgz",
        "input": "news-commentary-v12.de-en.en",
        "target": "news-commentary-v12.de-en.de",
    },
    {
        "url": "http://www.statmt.org/wmt13/training-parallel-commoncrawl.tgz",
        "input": "commoncrawl.de-en.en",
        "target": "commoncrawl.de-en.de",
    },
    {
        "url": "http://www.statmt.org/wmt13/training-parallel-europarl-v7.tgz",
        "input": "europarl-v7.de-en.en",
        "target": "europarl-v7.de-en.de",
    },
]
# Use pre-defined minimum count to generate subtoken vocabulary.
_TRAIN_DATA_MIN_COUNT = 6

_EVAL_DATA_SOURCES = [
    {
        "url": "http://data.statmt.org/wmt17/translation-task/dev.tgz",
        "input": "newstest2013.en",
        "target": "newstest2013.de",
    }
]

# Vocabulary constants
_REFERENCE_DICT_FILE = "reference_dictionary.ende.txt"
_TARGET_VOCAB_SIZE = 32768  # Number of subtokens in the vocabulary list.
_TARGET_THRESHOLD = 327  # Accept vocabulary if size is within this threshold
_VOCAB_FILE = "dict.en.txt"

# Strings to inclue in the generated files.
_PREFIX = "wmt32k"
_COMPILE_TAG = "compiled"
_ENCODE_TAG = ""
_TRAIN_TAG = "train"
_EVAL_TAG = "dev"  # Following WMT and Tensor2Tensor conventions, in which the
# evaluation datasets are tagged as "dev" for development.
_TEST_TAG = "test"

# Number of files to split train and evaluation data
_TRAIN_SHARDS = 1
_EVAL_SHARDS = 1
_TEST_SHARDS = 1


def find_file(path, filename, max_depth=5):
    """Returns full filepath if the file is in path or a subdirectory."""
    for root, dirs, files in os.walk(path):
        if filename in files:
            return os.path.join(root, filename)

        # Don't search past max_depth
        depth = root[len(path) + 1 :].count(os.sep)
        if depth > max_depth:
            del dirs[:]  # Clear dirs
    return None


###############################################################################
# Download and extraction functions
###############################################################################
def get_raw_files(raw_dir, data_source):
    """Return raw files from source. Downloads/extracts if needed.

    Args:
      raw_dir: string directory to store raw files
      data_source: dictionary with
        {"url": url of compressed dataset containing input and target files
         "input": file with data in input language
         "target": file with data in target language}

    Returns:
      dictionary with
        {"inputs": list of files containing data in input language
         "targets": list of files containing corresponding data in target language
        }
    """
    raw_files = {
        "inputs": [],
        "targets": [],
    }  # keys
    for d in data_source:
        input_file, target_file = download_and_extract(
            raw_dir, d["url"], d["input"], d["target"]
        )
        raw_files["inputs"].append(input_file)
        raw_files["targets"].append(target_file)
    return raw_files


def download_report_hook(count, block_size, total_size):
    """Report hook for download progress.

    Args:
      count: current block number
      block_size: block size
      total_size: total size
    """
    percent = int(count * block_size * 100 / total_size)
    print("\r%d%%" % percent + " completed", end="\r")


def download_from_url(path, url):
    """Download content from a url.

    Args:
      path: string directory where file will be downloaded
      url: string url

    Returns:
      Full path to downloaded file
    """
    filename = url.split("/")[-1]
    found_file = find_file(path, filename, max_depth=0)
    if found_file is None:
        filename = os.path.join(path, filename)
        print("Downloading from %s to %s." % (url, filename))
        inprogress_filepath = filename + ".incomplete"
        inprogress_filepath, _ = urllib.request.urlretrieve(
            url, inprogress_filepath, reporthook=download_report_hook
        )
        # Print newline to clear the carriage return from the download progress.
        print()
        os.rename(inprogress_filepath, filename)
        return filename
    else:
        print("Already downloaded: %s (at %s)." % (url, found_file))
        return found_file


def download_and_extract(path, url, input_filename, target_filename):
    """Extract files from downloaded compressed archive file.

    Args:
      path: string directory where the files will be downloaded
      url: url containing the compressed input and target files
      input_filename: name of file containing data in source language
      target_filename: name of file containing data in target language

    Returns:
      Full paths to extracted input and target files.

    Raises:
      OSError: if the the download/extraction fails.
    """
    # Check if extracted files already exist in path
    input_file = find_file(path, input_filename)
    target_file = find_file(path, target_filename)
    if input_file and target_file:
        print("Already downloaded and extracted %s." % url)
        return input_file, target_file

    # Download archive file if it doesn't already exist.
    compressed_file = download_from_url(path, url)

    # Extract compressed files
    print("Extracting %s." % compressed_file)
    with tarfile.open(compressed_file, "r:gz") as corpus_tar:
        corpus_tar.extractall(path)

    # Return filepaths of the requested files.
    input_file = find_file(path, input_filename)
    target_file = find_file(path, target_filename)

    if input_file and target_file:
        return input_file, target_file

    raise OSError("Download/extraction failed for url %s to path %s" % (url, path))


def txt_line_iterator(path):
    """Iterate through lines of file."""
    with open(path, mode="r", newline="\n") as f:
        for line in f:
            yield line.strip()


def compile_files(data_dir, raw_files, tag):
    """Compile raw files into a single file for each language.

    Args:
      raw_dir: Directory containing downloaded raw files.
      raw_files: Dict containing filenames of input and target data.
        {"inputs": list of files containing data in input language
         "targets": list of files containing corresponding data in target language
        }
      tag: String to append to the compiled filename.

    Returns:
      Full path of compiled input and target files.
    """
    print("Compiling files with tag %s." % tag)
    filename = "%s-%s-%s" % (_PREFIX, _COMPILE_TAG, tag)
    input_compiled_file = os.path.join(data_dir, filename + ".lang1")
    target_compiled_file = os.path.join(data_dir, filename + ".lang2")

    with open(input_compiled_file, mode="w", newline="\n") as input_writer:
        with open(target_compiled_file, mode="w", newline="\n") as target_writer:
            for i in range(len(raw_files["inputs"])):
                input_file = raw_files["inputs"][i]
                target_file = raw_files["targets"][i]

                print("Reading files %s and %s." % (input_file, target_file))
                write_file(input_writer, input_file)
                write_file(target_writer, target_file)
    return input_compiled_file, target_compiled_file


def write_file(writer, filename):
    """Write all of lines from file using the writer."""
    for line in txt_line_iterator(filename):
        writer.write(line)
        writer.write("\n")


###############################################################################
# Data preprocessing
###############################################################################
def encode_and_save_files_utf8(subtokenizer, data_dir, raw_files, tag, total_shards):
    """Save data from files as encoded example pairs in UT8 format.

    Args:
      subtokenizer: Subtokenizer object that will be used to encode the strings.
      data_dir: The directory in which to write the examples
      raw_files: A tuple of (input, target) data files. Each line in the input and
        the corresponding line in target file will be saved in encoded format (vocab integer ids).
      tag: String that will be added onto the file names.
      total_shards: Number of files to divide the data into.

    Returns:
      List of all files produced.
    """
    # Create a file for each shard.
    if total_shards > 1:
        filepaths = [
            shard_filename(data_dir + "/utf8", tag, n + 1, total_shards)
            for n in range(total_shards)
        ]
    else:
        filepaths = [
            shard_filename_single(data_dir + "/utf8", tag) for n in range(total_shards)
        ]

    if all_exist(filepaths):
        print("Files with tag %s already exist." % tag)
        return filepaths

    print("Saving files with tag %s." % tag)
    input_file = raw_files[0]
    target_file = raw_files[1]

    # Write examples to each shard in round robin order.
    tmp_filepaths = [fname + ".incomplete" for fname in filepaths]
    src_writers = [
        open(fname + ".en", mode="w", newline="\n") for fname in tmp_filepaths
    ]
    dst_writers = [
        open(fname + ".de", mode="w", newline="\n") for fname in tmp_filepaths
    ]

    counter, shard_idx = 0, 0
    src_sum, tgt_sum = 0, 0
    for counter, (input_line, target_line) in enumerate(
        zip(txt_line_iterator(input_file), txt_line_iterator(target_file))
    ):
        if counter > 0 and counter % 100000 == 0:
            print("\tSaving case %d." % counter)

        src_sum += len(input_line.split(" "))
        tgt_sum += len(target_line.split(" "))

        src_writers[shard_idx].write(
            " ".join(
                [
                    str(idx)
                    for idx in subtokenizer.encode(input_line.strip(), add_eos=True)
                ]
            )
        )
        dst_writers[shard_idx].write(
            " ".join(
                [
                    str(idx)
                    for idx in subtokenizer.encode(target_line.strip(), add_eos=True)
                ]
            )
        )

        src_writers[shard_idx].write("\n")
        dst_writers[shard_idx].write("\n")

        shard_idx = (shard_idx + 1) % total_shards

    for writer in src_writers:
        writer.close()

    for writer in dst_writers:
        writer.close()

    for tmp_name, final_name in zip(tmp_filepaths, filepaths):
        src_tmp_name = tmp_name + ".en"
        dst_tmp_name = tmp_name + ".de"

        src_fnl_name = final_name + ".en"
        dst_fnl_name = final_name + ".de"

        os.rename(src_tmp_name, src_fnl_name)
        os.rename(dst_tmp_name, dst_fnl_name)

    print("Saved %d Examples", counter)

    joined_file_list = []
    for file in filepaths:
        joined_file_list.append(file + ".en")
        joined_file_list.append(file + ".de")

    return joined_file_list


def shard_filename_single(path, tag):
    """Create filename for data shard."""
    return os.path.join(path, "%s.en-de" % (tag))


def shard_filename(path, tag, shard_num, total_shards):
    """Create filename for data shard."""
    return os.path.join(path, "%s%d.en-de" % (tag, shard_num))


def shuffle_records(fname):
    """Shuffle records in a single file."""
    print("Shuffling records.")
    src = "en-de.en"
    dst = "en-de.de"
    file_sets = {
        "train": {src: [], dst: []},
        "dev": {src: [], dst: []},
        "test": {src: [], dst: []},
    }

    for f in fname:
        tmp_fname = f + ".unshuffled"
        os.rename(f, tmp_fname)
        for split in file_sets:
            if split in f:
                print("fname:", tmp_fname)
                if src in f:
                    file_sets[split][src].append(tmp_fname)
                else:
                    file_sets[split][dst].append(tmp_fname)

    for split in file_sets:
        if file_sets[split][src]:
            file_sets[split][src].sort

        if file_sets[split][dst]:
            file_sets[split][dst].sort

    print("file_set:", file_sets)

    for split in file_sets:
        print("split:", split)
        if file_sets[split][src] and file_sets[split][dst]:
            for fs, fd in zip(file_sets[split][src], file_sets[split][dst]):
                print("zipped fs, fd")
                with open(fs, mode="r", newline="\n") as src_reader:
                    with open(fd, mode="r", newline="\n") as tgt_reader:
                        print("opened for write:", fs, fd)
                        records = []
                        for rs, rd in zip(src_reader, tgt_reader):
                            records.append((rs, rd))
                            if len(records) % 100000 == 0:
                                print("\t Read : %d", len(records))

                print("performing shuffle")
                random.shuffle(records)

                with open(fs[:-11], mode="w", newline="\n") as src_writer:
                    with open(fd[:-11], mode="w", newline="\n") as tgt_writer:
                        print("opened for write:", fs, fd)
                        for i, record in enumerate(records):
                            src_writer.write(record[0])
                            tgt_writer.write(record[1])
                            if i > 0 and i % 100000 == 0:
                                print("\tWriting record: %d" % i)

                os.remove(fs)
                os.remove(fd)


def all_exist(filepaths):
    """Returns true if all files in the list exist."""
    for fname in filepaths:
        if not os.path.exists(fname):
            return False
    return True


def make_dir(path):
    if not os.path.exists(path):
        print("Creating directory %s" % path)
        os.mkdir(path)


def mock_binarize(filename, consumer):
    nseq, ntok = 0, 0
    replaced = Counter()

    with open(filename, mode="r", newline="\n") as f:
        for line in f:
            nwords = len(line.split(" "))
            ids = torch.IntTensor(nwords).long()
            for idx, tok in enumerate(line.split(" ")):
                ids[idx] = int(tok)

            nseq += 1

            consumer(ids)
            ntok += len(ids)

    return {
        "nseq": nseq,
        "nunk": sum(replaced.values()),
        "ntok": ntok,
        "replaced": len(replaced),
    }


def make_binary_dataset(data_dir):
    files = glob.glob(data_dir + "/utf8/test*.en", recursive=True)
    files += glob.glob(data_dir + "/utf8/test*.de", recursive=True)

    files += glob.glob(data_dir + "/utf8/dev*.en", recursive=True)
    files += glob.glob(data_dir + "/utf8/dev*.de", recursive=True)

    files += glob.glob(data_dir + "/utf8/train*.en", recursive=True)
    files += glob.glob(data_dir + "/utf8/train*.de", recursive=True)

    def consumer(tensor):
        ds.add_item(tensor)

    for file in files:
        print("Converting file:", file)
        ds = indexed_dataset.IndexedDatasetBuilder(file + ".bin")

        res = mock_binarize(file, consumer)

        ds.finalize(file + ".idx")


def main():
    """Obtain training and evaluation data for the Transformer model."""
    print("Step 1/6: Create directories and copy dictionary")
    make_dir(FLAGS.raw_dir)
    make_dir(FLAGS.data_dir)
    en_dict, de_dict = "dict.en.txt", "dict.de.txt"

    curr_dir = os.path.dirname(os.path.abspath(__file__))

    en_dict_path = os.path.abspath(os.path.join(FLAGS.data_dir, en_dict))
    de_dict_path = os.path.abspath(os.path.join(FLAGS.data_dir, de_dict))
    shutil.copyfile(os.path.join(curr_dir, _REFERENCE_DICT_FILE), en_dict_path)
    shutil.copyfile(os.path.join(curr_dir, _REFERENCE_DICT_FILE), de_dict_path)

    cmd = "\"1s/^/'<lua_index_compat>'\\n/\""
    subprocess.call(["sed -i {} {}".format(cmd, en_dict_path)], shell=True)
    subprocess.call(["sed -i {} {}".format(cmd, de_dict_path)], shell=True)

    utf8_path = os.path.join(FLAGS.data_dir, "utf8")
    make_dir(utf8_path)
    shutil.copyfile(
        os.path.join(curr_dir, _REFERENCE_DICT_FILE), os.path.join(utf8_path, en_dict)
    )
    shutil.copyfile(
        os.path.join(curr_dir, _REFERENCE_DICT_FILE), os.path.join(utf8_path, de_dict)
    )

    # Copy test files to data_dir
    test_files = [
        os.path.join(curr_dir, "newstest2014.en"),
        os.path.join(curr_dir, "newstest2014.de"),
    ]

    for t in test_files:
        shutil.copy(t, FLAGS.data_dir)

    # Get paths of download/extracted training and evaluation files.
    print("Step 2/6: Downloading data from source")
    train_files = get_raw_files(FLAGS.raw_dir, _TRAIN_DATA_SOURCES)
    eval_files = get_raw_files(FLAGS.raw_dir, _EVAL_DATA_SOURCES)

    test_files = [
        os.path.join(FLAGS.data_dir, "newstest2014.en"),
        os.path.join(FLAGS.data_dir, "newstest2014.de"),
    ]

    # Create subtokenizer based on the training files.
    print("Step 3/6: Creating subtokenizer and building vocabulary")
    # train_files_flat = train_files["inputs"] + train_files["targets"]
    vocab_file = os.path.join(FLAGS.data_dir, _VOCAB_FILE)
    subtokenizer = sub_tokenizer.Subtokenizer.init_from_existing_vocab_file(vocab_file)

    print("Step 4/6: Compiling training and evaluation data")
    compiled_train_files = compile_files(FLAGS.data_dir, train_files, _TRAIN_TAG)
    compiled_eval_files = compile_files(FLAGS.data_dir, eval_files, _EVAL_TAG)
    compiled_test_files = test_files

    # Tokenize and save data as example pairs in UTF8 format.
    print("Step 5/6: Preprocessing and saving data")
    train_utf8_files = encode_and_save_files_utf8(
        subtokenizer, FLAGS.data_dir, compiled_train_files, _TRAIN_TAG, _TRAIN_SHARDS
    )
    encode_and_save_files_utf8(
        subtokenizer, FLAGS.data_dir, compiled_eval_files, _EVAL_TAG, _EVAL_SHARDS
    )
    encode_and_save_files_utf8(
        subtokenizer, FLAGS.data_dir, compiled_test_files, _TEST_TAG, _TEST_SHARDS
    )

    shuffle_records(train_utf8_files)

    # Convert to binary
    print("Step 6/6: Binarizing dataset")
    make_binary_dataset(FLAGS.data_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-dir",
        "-dd",
        type=str,
        default="/tmp/translate_ende",
        help="[default: %(default)s] Directory for where the "
        "translate_ende_wmt32k dataset is saved.",
        metavar="<DD>",
    )
    parser.add_argument(
        "--raw-dir",
        "-rd",
        type=str,
        default="/tmp/translate_ende_raw",
        help="[default: %(default)s] Path where the raw data will be downloaded "
        "and extracted.",
        metavar="<RD>",
    )
    parser.add_argument(
        "--search",
        action="store_true",
        help="If set, use binary search to find the vocabulary set with size"
        "closest to the target size (%d)." % _TARGET_VOCAB_SIZE,
    )

    FLAGS, unparsed = parser.parse_known_args()
    main()



import torch


@torch.jit.script
def orthogonalize(matrix, eps=torch.FloatTensor([1e-16])):
    """Function used to orthogonalize a matrix.

    Args:
        matrix (torch.Tensor): Matrix to orthogonalize
        eps (torch.FloatTensor): Used to avoid division by zero (default: 1e-16)
    """
    n, m = matrix.shape
    for i in range(m):
        # Normalize the i'th column
        col = matrix[:, i : i + 1]
        col /= torch.sqrt(torch.sum(col ** 2)) + eps
        # Project it on the rest and remove it
        if i + 1 < m:
            rest = matrix[:, i + 1 :]
            # rest -= torch.matmul(col.t(), rest) * col
            rest -= torch.sum(col * rest, dim=0) * col



import torch


class Dictionary(object):
    """Dictionary Class for WMT17 Dataset.
    Essentially a mapping from symbols to consecutive integers

    Args:
        pad (str): Padding symbol to use
        eos (str): End of String symbol to use
    """

    def __init__(self, pad="<pad>_", eos="<EOS>_"):
        self.pad_word, self.eos_word = pad, eos
        self.symbols = []
        self.count = []
        self.indices = {}
        # dictionary indexing starts at 1 for consistency with Lua
        # Commented out and hard-coded since pad and eos are in the dictionary files already
        self.add_symbol("<lua_index_compat>")
        self.pad_index = 1
        self.eos_index = 2
        self.nspecial = 3

    def __eq__(self, other):
        return self.indices == other.indices

    def __getitem__(self, idx):
        if idx < len(self.symbols):
            return self.symbols[idx]
        else:
            assert idx < len(self.symbols)

    def __len__(self):
        """Returns the number of symbols in the dictionary"""
        return len(self.symbols)

    def index(self, sym):
        """Returns the index of the specified symbol"""
        if sym in self.indices:
            return self.indices[sym]
        else:
            assert sym in self.indices

    def string(self, tensor, bpe_symbol=None):
        """Helper for converting a tensor of token indices to a string.

        Can optionally remove BPE symbols or escape <unk> words.
        """
        if torch.is_tensor(tensor) and tensor.dim() == 2:
            return "\n".join(self.string(t) for t in tensor)

        def token_string(i):
            return self[i]

        sent = " ".join(token_string(i) for i in tensor if i != self.eos())
        if bpe_symbol is not None:
            sent = (sent + " ").replace(bpe_symbol, "").rstrip()

        return sent

    def add_symbol(self, word, n=1):
        """Adds a word to the dictionary"""
        if word in self.indices:
            idx = self.indices[word]
            self.count[idx] = self.count[idx] + n
            return idx
        else:
            idx = len(self.symbols)
            self.indices[word] = idx
            self.symbols.append(word)
            self.count.append(n)
            return idx

    def update(self, new_dict):
        """Updates counts from new dictionary."""
        for word in new_dict.symbols:
            idx2 = new_dict.indices[word]
            if word in self.indices:
                idx = self.indices[word]
                self.count[idx] = self.count[idx] + new_dict.count[idx2]
            else:
                idx = len(self.symbols)
                self.indices[word] = idx
                self.symbols.append(word)
                self.count.append(new_dict.count[idx2])

    def pad(self):
        """Helper to get index of pad symbol"""
        return self.pad_index

    def eos(self):
        """Helper to get index of end-of-sentence symbol"""
        return self.eos_index

    @classmethod
    def load(cls, f, ignore_utf_errors=False):
        """Loads the dictionary from a text file with the format:

        ```
        <symbol0>
        <symbol1>
        ...
        ```

        Args:
            f (str): Dictionary file name
            ignore_utf_errors (bool): Ignore UTF-8 related errors
        """
        if isinstance(f, str):
            try:
                if not ignore_utf_errors:
                    with open(f, "r", encoding="utf-8") as fd:
                        return cls.load(fd)
                else:
                    with open(f, "r", encoding="utf-8", errors="ignore") as fd:
                        return cls.load(fd)

            except FileNotFoundError as fnfe:
                raise fnfe

            except Exception:
                raise Exception(
                    "Incorrect encoding detected in {}, please rebuild the dataset".format(
                        f
                    )
                )

        d = cls()
        for line in f.readlines():
            word = line.strip()[1:-1]
            count = 1
            d.indices[word] = len(d.symbols)
            d.symbols.append(word)
            d.count.append(count)

        n_pad_tokens_on_end = 33712 - len(d.symbols)

        for i in range(n_pad_tokens_on_end):
            pad_str = "<pad000" + str(i) + ">"
            d.indices[pad_str] = len(d.symbols)
            d.symbols.append(pad_str)
            d.count.append(1)

        return d



# Copyright 2018 MLBenchmark Group. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Defines Subtokenizer class to encode and decode strings."""

import os
import re
import sys
import unicodedata

import numpy as np

# from __future__ import absolute_import
# from __future__ import division
# from __future__ import print_function
import six
from six.moves import xrange  # pylint: disable=redefined-builtin

LUA = "<lua_index_compat>"
PAD = "<pad>_"
PAD_ID = 1
EOS = "<EOS>_"
EOS_ID = 2
UNK = "<bypass_unk>"
RESERVED_TOKENS = [LUA, PAD, EOS, UNK]

# Set of characters that will be used in the function _escape_token() (see func
# docstring for more details).
# This set is added to the alphabet list to ensure that all escaped tokens can
# be encoded.
_ESCAPE_CHARS = set(u"\\_u;0123456789")
# Regex for the function _unescape_token(), the inverse of _escape_token().
# This is used to find "\u", "\\", and "\###;" substrings in the token.
_UNESCAPE_REGEX = re.compile(r"\\u|\\\\|\\([0-9]+);")
#
_UNDEFINED_UNICODE = u"\u3013"
#
# # Set contains all letter and number characters.
_ALPHANUMERIC_CHAR_SET = set(
    six.unichr(i)
    for i in xrange(sys.maxunicode)
    if (
        unicodedata.category(six.unichr(i)).startswith("L")
        or unicodedata.category(six.unichr(i)).startswith("N")
    )
)


class Subtokenizer(object):
    """Encodes and decodes strings to/from integer IDs."""

    def __init__(self, vocab_file, reserved_tokens=None):
        """Initializes class, creating a vocab file if data_files is provided."""
        print("Initializing Subtokenizer from file %s." % vocab_file)

        if reserved_tokens is None:
            reserved_tokens = RESERVED_TOKENS
        elif reserved_tokens is "assumed_in_file":
            reserved_tokens = []

        self.subtoken_list = _load_vocab_file(vocab_file, reserved_tokens)
        self.alphabet = _generate_alphabet_dict(self.subtoken_list, reserved_tokens)
        self.subtoken_to_id_dict = _list_to_index_dict(self.subtoken_list)

        self.max_subtoken_length = 0
        for subtoken in self.subtoken_list:
            self.max_subtoken_length = max(self.max_subtoken_length, len(subtoken))

        # Create cache to speed up subtokenization
        self._cache_size = 2 ** 20
        self._cache = [(None, None)] * self._cache_size

    @staticmethod
    def init_from_existing_vocab_file(vocab_file):
        """Create subtoken vocabulary based on files, and save vocab to file.

        Args:
          vocab_file: String name of vocab file to store subtoken vocabulary.

        Returns:
          Subtokenizer object
        """

        if os.path.exists(vocab_file):
            print("Vocab file exists (%s)" % vocab_file)
        else:
            print("Vocab file does not exist (%s)" % vocab_file)

        return Subtokenizer(vocab_file, reserved_tokens="assumed_in_file")

    def encode(self, raw_string, add_eos=False):
        """Encodes a string into a list of int subtoken ids."""
        ret = []
        tokens = _split_string_to_tokens(raw_string)
        for token in tokens:
            ret.extend(self._token_to_subtoken_ids(token))
        if add_eos:
            ret.append(EOS_ID)
        return ret

    def _token_to_subtoken_ids(self, token):
        """Encode a single token into a list of subtoken ids."""
        cache_location = hash(token) % self._cache_size
        cache_key, cache_value = self._cache[cache_location]
        if cache_key == token:
            return cache_value

        ret = _split_token_to_subtokens(
            _escape_token(token, self.alphabet),
            self.subtoken_to_id_dict,
            self.max_subtoken_length,
        )
        ret = [self.subtoken_to_id_dict[subtoken_id] for subtoken_id in ret]

        self._cache[cache_location] = (token, ret)
        return ret

    def decode(self, subtokens):
        """Converts list of int subtokens ids into a string."""
        if isinstance(subtokens, np.ndarray):
            # Note that list(subtokens) converts subtokens to a python list, but the
            # items remain as np.int32. This converts both the array and its items.
            subtokens = subtokens.tolist()

        if not subtokens:
            return ""

        assert isinstance(subtokens, list) and isinstance(
            subtokens[0], int
        ), "Subtokens argument passed into decode() must be a list of integers."

        return _join_tokens_to_string(self._subtoken_ids_to_tokens(subtokens))

    def _subtoken_ids_to_tokens(self, subtokens):
        """Convert list of int subtoken ids to a list of string tokens."""
        escaped_tokens = "".join(
            [self.subtoken_list[s] for s in subtokens if s < len(self.subtoken_list)]
        )
        escaped_tokens = escaped_tokens.split("_")

        # All tokens in the vocabulary list have been escaped (see _escape_token())
        # so each token must be unescaped when decoding.
        ret = []
        for token in escaped_tokens:
            if token:
                ret.append(_unescape_token(token))
        return ret


def _load_vocab_file(vocab_file, reserved_tokens=None):
    """Load vocabulary while ensuring reserved tokens are at the top."""
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS

    subtoken_list = []
    with open(vocab_file, mode="r", newline="\n") as f:
        for line in f:
            subtoken = line.strip()
            subtoken = subtoken[1:-1]  # Remove surrounding single-quotes
            if subtoken in reserved_tokens:
                continue
            subtoken_list.append(subtoken)
    return reserved_tokens + subtoken_list


def _split_string_to_tokens(text):
    """Splits text to a list of string tokens."""
    if not text:
        return []
    ret = []
    token_start = 0
    # Classify each character in the input string
    is_alnum = [c in _ALPHANUMERIC_CHAR_SET for c in text]
    for pos in xrange(1, len(text)):
        if is_alnum[pos] != is_alnum[pos - 1]:
            token = text[token_start:pos]
            if token != u" " or token_start == 0:
                ret.append(token)
            token_start = pos
    final_token = text[token_start:]
    ret.append(final_token)
    return ret


def _join_tokens_to_string(tokens):
    """Join a list of string tokens into a single string."""
    token_is_alnum = [t[0] in _ALPHANUMERIC_CHAR_SET for t in tokens]
    ret = []
    for i, token in enumerate(tokens):
        if i > 0 and token_is_alnum[i - 1] and token_is_alnum[i]:
            ret.append(u" ")
        ret.append(token)
    return "".join(ret)


def _escape_token(token, alphabet):
    r"""Replace characters that aren't in the alphabet and append "_" to token.

    Apply three transformations to the token:
      1. Replace underline character "_" with "\u", and backslash "\" with "\\".
      2. Replace characters outside of the alphabet with "\###;", where ### is the
         character's Unicode code point.
      3. Appends "_" to mark the end of a token.

    Args:
      token: unicode string to be escaped
      alphabet: list of all known characters

    Returns:
      escaped string
    """
    token = token.replace(u"\\", u"\\\\").replace(u"_", u"\\u")
    ret = [c if c in alphabet and c != u"\n" else r"\%d;" % ord(c) for c in token]
    return u"".join(ret) + "_"


def _unescape_token(token):
    r"""Replaces escaped characters in the token with their unescaped versions.

    Applies inverse transformations as _escape_token():
      1. Replace "\u" with "_", and "\\" with "\".
      2. Replace "\###;" with the unicode character the ### refers to.

    Args:
      token: escaped string

    Returns:
      unescaped string
    """

    def match(m):
        r"""Returns replacement string for matched object.

        Matched objects contain one of the strings that matches the regex pattern:
          r"\\u|\\\\|\\([0-9]+);"
        The strings can be '\u', '\\', or '\###;' (### is any digit number).

        m.group(0) refers to the entire matched string ('\u', '\\', or '\###;').
        m.group(1) refers to the first parenthesized subgroup ('###').

        m.group(0) exists for all match objects, while m.group(1) exists only for
        the string '\###;'.

        This function looks to see if m.group(1) exists. If it doesn't, then the
        matched string must be '\u' or '\\' . In this case, the corresponding
        replacement ('_' and '\') are returned. Note that in python, a single
        backslash is written as '\\', and double backslash as '\\\\'.

        If m.goup(1) exists, then use the integer in m.group(1) to return a
        unicode character.

        Args:
          m: match object

        Returns:
          String to replace matched object with.
        """
        # Check if the matched strings are '\u' or '\\'.
        if m.group(1) is None:
            return u"_" if m.group(0) == u"\\u" else u"\\"

        # If m.group(1) exists, try and return unicode character.
        try:
            return six.unichr(int(m.group(1)))
        except (ValueError, OverflowError) as _:
            return _UNDEFINED_UNICODE

    # Use match function to replace escaped substrings in the token.
    return _UNESCAPE_REGEX.sub(match, token)


def _list_to_index_dict(lst):
    """Create dictionary mapping list items to their indices in the list."""
    return {item: n for n, item in enumerate(lst)}


#
#
def _split_token_to_subtokens(token, subtoken_dict, max_subtoken_length):
    """Splits a token into subtokens defined in the subtoken dict."""
    ret = []
    start = 0
    token_len = len(token)
    while start < token_len:
        # Find the longest subtoken, so iterate backwards.
        for end in xrange(min(token_len, start + max_subtoken_length), start, -1):
            subtoken = token[start:end]
            if subtoken in subtoken_dict:
                ret.append(subtoken)
                start = end
                break
        else:  # Did not break
            # If there is no possible encoding of the escaped token then one of the
            # characters in the token is not in the alphabet. This should be
            # impossible and would be indicative of a bug.
            raise ValueError('Was unable to split token "%s" into subtokens.' % token)
    return ret


def _generate_alphabet_dict(iterable, reserved_tokens=None):
    """Create set of characters that appear in any element in the iterable."""
    if reserved_tokens is None:
        reserved_tokens = RESERVED_TOKENS
    elif reserved_tokens is "assumed_in_file":
        reserved_tokens = []

    alphabet = {c for token in iterable for c in token}
    alphabet |= {c for token in reserved_tokens for c in token}
    alphabet |= _ESCAPE_CHARS  # Add escape characters to alphabet set.

    return alphabet



import math

import torch
from torch import nn


def _make_positions(tensor, padding_idx, left_pad):
    """Replace non-padding symbols with their position numbers.

    Position numbers begin at padding_idx+1.

    Padding symbols are ignored, but it is necessary to specify whether padding
    is added on the left side (left_pad=True) or right side (left_pad=False).
    """
    max_pos = padding_idx + 1 + tensor.size(1)
    if not hasattr(_make_positions, "range_buf"):
        _make_positions.range_buf = tensor.new()
    _make_positions.range_buf = _make_positions.range_buf.type_as(tensor)
    if _make_positions.range_buf.numel() < max_pos:
        torch.arange(padding_idx + 1, max_pos, out=_make_positions.range_buf)
    mask = tensor.ne(padding_idx)
    positions = _make_positions.range_buf[: tensor.size(1)].expand_as(tensor)
    if left_pad:
        positions = positions - mask.size(1) + mask.long().sum(dim=1).unsqueeze(1)
    return tensor.clone().masked_scatter_(mask, positions[mask])


class LearnedPositionalEmbedding(nn.Embedding):
    """This module learns positional embeddings up to a fixed maximum size.

    Padding symbols are ignored, but it is necessary to specify whether padding
    is added on the left side (left_pad=True) or right side (left_pad=False).
    """

    def __init__(self, num_embeddings, embedding_dim, padding_idx, left_pad):
        super().__init__(num_embeddings, embedding_dim, padding_idx)
        self.left_pad = left_pad

    def forward(self, input, incremental_state=None):
        """Input is expected to be of size [bsz x seqlen]."""
        if incremental_state is not None:
            # positions is the same for every token when decoding a single step
            positions = input.data.new(1, 1).fill_(self.padding_idx + input.size(1))
        else:
            positions = _make_positions(input.data, self.padding_idx, self.left_pad)
        return super().forward(positions)

    def max_positions(self):
        """Maximum number of supported positions."""
        return self.num_embeddings - self.padding_idx - 1


class SinusoidalPositionalEmbedding(nn.Module):
    """This module produces sinusoidal positional embeddings of any length.

    Padding symbols are ignored, but it is necessary to specify whether padding
    is added on the left side (left_pad=True) or right side (left_pad=False).
    """

    def __init__(self, embedding_dim, padding_idx, left_pad, init_size=1024):
        super().__init__()
        self.embedding_dim = embedding_dim
        self.padding_idx = padding_idx
        self.left_pad = left_pad
        self.weights = SinusoidalPositionalEmbedding.get_embedding(
            init_size, embedding_dim, padding_idx,
        )
        self.register_buffer("_float_tensor", torch.FloatTensor(1))

    @staticmethod
    def get_embedding(num_embeddings, embedding_dim, padding_idx=None):
        """Build sinusoidal embeddings.

        This matches the implementation in tensor2tensor, but differs slightly
        from the description in Section 3.5 of "Attention Is All You Need".
        """
        half_dim = embedding_dim // 2
        emb = math.log(10000) / (half_dim - 1)
        emb = torch.exp(torch.arange(half_dim, dtype=torch.float) * -emb)
        emb = torch.arange(num_embeddings, dtype=torch.float).unsqueeze(
            1
        ) * emb.unsqueeze(0)
        emb = torch.cat([torch.sin(emb), torch.cos(emb)], dim=1).view(
            num_embeddings, -1
        )
        if embedding_dim % 2 == 1:
            # zero pad
            emb = torch.cat([emb, torch.zeros(num_embeddings, 1)], dim=1)
        if padding_idx is not None:
            emb[padding_idx, :] = 0
        return emb

    def forward(self, input, incremental_state=None):
        """Input is expected to be of size [bsz x seqlen]."""
        # recompute/expand embeddings if needed
        bsz, seq_len = input.size()
        max_pos = self.padding_idx + 1 + seq_len
        if self.weights is None or max_pos > self.weights.size(0):
            self.weights = SinusoidalPositionalEmbedding.get_embedding(
                max_pos, self.embedding_dim, self.padding_idx,
            )
        self.weights = self.weights.type_as(self._float_tensor)

        if incremental_state is not None:
            # positions is the same for every token when decoding a single step
            return self.weights[self.padding_idx + seq_len, :].expand(bsz, 1, -1)

        positions = _make_positions(input.data, self.padding_idx, self.left_pad)
        return (
            self.weights.index_select(0, positions.view(-1))
            .view(bsz, seq_len, -1)
            .detach()
        )

    def max_positions(self):
        """Maximum number of supported positions."""
        return int(1e5)  # an arbitrary large number


def PositionalEmbedding(
    num_embeddings, embedding_dim, padding_idx, left_pad, learned=False
):
    if learned:
        m = LearnedPositionalEmbedding(
            num_embeddings + padding_idx + 1, embedding_dim, padding_idx, left_pad
        )
        nn.init.normal_(m.weight, mean=0, std=embedding_dim ** -0.5)
        nn.init.constant_(m.weight[padding_idx], 0)
    else:
        m = SinusoidalPositionalEmbedding(
            embedding_dim, padding_idx, left_pad, num_embeddings + padding_idx + 1
        )
    return m


def _parse_embedding(embed_path):
    """Parse embedding text file into a dictionary of word and embedding tensors.

    The first line can have vocabulary size and dimension. The following lines
    should contain word and embedding separated by spaces.

    Example:
        2 5
        the -0.0230 -0.0264  0.0287  0.0171  0.1403
        at -0.0395 -0.1286  0.0275  0.0254 -0.0932
    """
    embed_dict = {}
    with open(embed_path) as f_embed:
        next(f_embed)  # skip header
        for line in f_embed:
            pieces = line.rstrip().split(" ")
            embed_dict[pieces[0]] = torch.Tensor(
                [float(weight) for weight in pieces[1:]]
            )
    return embed_dict


def _load_embedding(embed_dict, vocab, embedding):
    for idx in range(len(vocab)):
        token = vocab[idx]
        if token in embed_dict:
            embedding.weight.data[idx] = embed_dict[token]
    return embedding


def build_embedding(dictionary, embed_dim, path=None):
    num_embeddings = len(dictionary)
    padding_idx = dictionary.pad()

    emb = nn.Embedding(num_embeddings, embed_dim, padding_idx=padding_idx)
    nn.init.normal_(emb.weight, mean=0, std=embed_dim ** -0.5)
    nn.init.constant_(emb.weight[padding_idx], 0)
    # if provided, load from preloaded dictionaries
    if path:
        embed_dict = _parse_embedding(path)
        _load_embedding(embed_dict, dictionary, emb)
    return emb



import os

from setuptools import setup

import mlbench_core

base__dir = os.path.dirname(mlbench_core.__file__)
ext_modules = []
try:
    from torch.utils.cpp_extension import BuildExtension, CUDAExtension, CUDA_HOME

    dir_ = os.path.join(
        base__dir, "models/pytorch/transformer/modules/strided_batched_gemm"
    )
    strided_batched_gemm = CUDAExtension(
        name="mlbench_core.models.pytorch.transformer.modules.strided_batched_gemm",
        sources=[
            os.path.join(dir_, "strided_batched_gemm.cpp"),
            os.path.join(dir_, "strided_batched_gemm_cuda.cu"),
        ],
        extra_compile_args={
            "cxx": ["-O2",],
            "nvcc": [
                "--gpu-architecture=compute_70",
                "--gpu-code=sm_70",
                "-O3",
                "-U__CUDA_NO_HALF_OPERATORS__",
                "-U__CUDA_NO_HALF_CONVERSIONS__",
            ],
        },
    )
    ext_modules.append(strided_batched_gemm)
    cmdclass = {"build_ext": BuildExtension}

    setup(ext_modules=ext_modules, cmdclass=cmdclass)

except (ImportError, OSError) as e:
    raise ValueError("Cannot install extensions because CUDA was not found")



import math

import torch
import torch.nn.functional as F
from torch import nn

from mlbench_core.models.pytorch.transformer.modules import (
    PositionalEmbedding,
    TransformerDecoderLayer,
)


class TransformerDecoder(nn.Module):
    """
    Transformer decoder consisting of *args.decoder_layers* layers. Each layer
    is a :class:`TransformerDecoderLayer`.

    Args:
        args: Arguments of model. All arguments should be accessible via `__getattribute__` method
        dictionary (`obj`:mlbench_core.dataset.nlp.pytorch.wmt17.Dictionary): decoding dictionary
        embed_tokens (torch.nn.Embedding): output embedding
        no_encoder_attn (bool, optional): whether to attend to encoder outputs
            (default: False).
        left_pad (bool): Pad targets to the left (`True`) or right (`False`). Default: `False`
    """

    def __init__(
        self, args, dictionary, embed_tokens, no_encoder_attn=False, left_pad=False
    ):
        super().__init__()
        self.dictionary = dictionary
        self.dropout = args.dropout
        self.share_input_output_embed = args.share_decoder_input_output_embed

        embed_dim = embed_tokens.embedding_dim
        padding_idx = embed_tokens.padding_idx
        self.max_target_positions = args.max_target_positions

        self.embed_tokens = embed_tokens
        self.embed_scale = math.sqrt(embed_dim)
        self.embed_positions = (
            PositionalEmbedding(
                args.max_target_positions,
                embed_dim,
                padding_idx,
                left_pad=left_pad,
                learned=args.decoder_learned_pos,
            )
            if not args.no_token_positional_embeddings
            else None
        )

        self.layers = nn.ModuleList(
            [
                TransformerDecoderLayer(args, no_encoder_attn)
                for _ in range(args.decoder_layers)
            ]
        )

        if not self.share_input_output_embed:
            self.embed_out = nn.Parameter(torch.Tensor(len(dictionary), embed_dim))
            nn.init.normal_(self.embed_out, mean=0, std=embed_dim ** -0.5)
        self.normalize = args.decoder_normalize_before

        if self.normalize:
            self.layer_norm = nn.LayerNorm(embed_dim)

    def forward(self, prev_output_tokens, encoder_out=None, incremental_state=None):
        # embed positions
        positions = (
            self.embed_positions(
                prev_output_tokens, incremental_state=incremental_state,
            )
            if self.embed_positions is not None
            else None
        )

        if incremental_state is not None:
            prev_output_tokens = prev_output_tokens[:, -1:]
            if positions is not None:
                positions = positions[:, -1:]

        # embed tokens and positions
        x = self.embed_scale * self.embed_tokens(prev_output_tokens)
        if positions is not None:
            x += positions
        x = F.dropout(x, p=self.dropout, training=self.training)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)
        attn = None

        # decoder layers
        for layer in self.layers:
            x, attn = layer(
                x,
                encoder_out["encoder_out"] if encoder_out is not None else None,
                encoder_out["encoder_padding_mask"]
                if encoder_out is not None
                else None,
                incremental_state,
            )

        if self.normalize:
            x = self.layer_norm(x)

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        # project back to size of vocabulary
        if self.share_input_output_embed:
            x = F.linear(x, self.embed_tokens.weight)
        else:
            x = F.linear(x, self.embed_out)

        return x, attn

    def max_positions(self):
        """Maximum output length supported by the decoder."""
        if self.embed_positions is None:
            return self.max_target_positions
        return min(self.max_target_positions, self.embed_positions.max_positions())

    def reorder_incremental_state(self, incremental_state, new_order):
        """Reorder incremental state.

        This should be called when the order of the input has changed from the
        previous time step. A typical use case is beam search, where the input
        order changes between time steps based on the selection of beams.
        """

        def apply_reorder_incremental_state(module):
            if module != self and hasattr(module, "reorder_incremental_state"):
                module.reorder_incremental_state(
                    incremental_state, new_order,
                )

        self.apply(apply_reorder_incremental_state)
